/* $RCSfile: $
 * $Revision: $ $Date: $
 * Auth: Jochen Fritz (jfritz@steptools.com)
 * 
 * 	Copyright (c) 1991-2013 by 
 * 	STEP Tools Inc., Troy, New York
 * 	All Rights Reserved
 * 
 * 	This software is furnished under a license and may be used and
 * 	copied only in accordance with the terms of such license and with
 * 	the inclusion of the above copyright notice.  This software and
 * 	accompanying written materials or any other copies thereof may
 * 	not be provided or otherwise made available to any other person.
 * 	No title to or ownership of the software is hereby transferred.
 * 
 * 		----------------------------------------
 */

"use strict";

P21.Lexer = (function() {


    var cls = function (filename) {
//	this.filename = filename;
//	var f = open_file_read(filename);

//	if (!f)
//	    throw new Error ("Could not open file: "+filename);
	
//	this.stream = f

	this.lineno = 1;
    }

    // P21 Character classes
    cls.EOF=0;
    cls.OOA=1;  // Out of Alphabet
    cls.SPACE=2;
    cls.DIGIT=3;
    cls.LOWER=4;
    cls.UPPER=5;
    cls.SPECIAL=6;
    cls.REVERSE_SOLIDUS=7;
    cls.APOSTROPHE=8;
    cls.HIGH_CODEPOINT=9;

    // Token values -- these are strings.
    cls.TOK_COMMENT   = "comment";
    cls.TOK_KEYWORD   = "keyword";
    cls.TOK_PRIMITIVE = "primitive";
    cls.TOK_NULL      = "null";
    cls.TOK_OMITTED   = "omitted";
    cls.TOK_ID        = "ID";
    cls.TOK_ENUM      = "enum";
    cls.TOK_STRING    = "string";
    cls.TOK_BINARY    = "binary";
    cls.TOK_URI       = "uri";
    cls.TOK_PUNCT     = "punct";
    cls.TOK_INTEGER   = "integer";
    cls.TOK_REAL      = "real";
    cls.TOK_NUMBER    = "number";
    
    cls.prototype.close = function() {
	close_file_read(this.stream);	
    }    
        
    function get_char_class(ch){
    
	if (!ch)
	    return cls.EOF;
	
	else if (ch == ' ')
	    return cls.SPACE;
    
	else if (ch >= '0' && ch <= '9')
	    return cls.DIGIT;
	
	else if (ch >= 'a' && ch <= 'z')
	    return cls.LOWER;
	
	else if (ch >= 'A' && ch <= 'Z' || ch == '_')
	    return cls.UPPER;
	
	else if (ch=='!' || ch=='"' || ch=='*' || ch=='$' || ch=='%' ||
		 ch=='&' || ch=='.' || ch=='#' || ch=='+' || ch==',' ||
		 ch=='-' || ch=='(' || ch==')' || ch=='?' || ch=='/' ||
		 ch==':' || ch=='<' || ch=='=' || ch=='>' || ch==';' ||
		 ch=='@' || ch=='[' || ch==']' || ch=='{' || ch=='|' ||
		 ch=='}' || ch=='^' || ch=='`' || ch=='~')
	    return cls.SPECIAL;
	
	else if (ch == '\\')
	    return cls.REVERSE_SOLIDUS;
	
	else if (ch == '\'')
	    return cls.APOSTROPHE;
	
	else if (ch > '\xff')
	    return cls.HIGH_CODEPOINT;
	
	return cls.OOA;
    }
    
    function scan_comment(lex, ch) {

	switch (lex.state) {
	    
	case null:
	    /* After initial /, but before * */
    	    if (ch != '*') {
    		throw new P21.SyntaxError (
		    lex, "have / that did not start a comment");
    	    }
	    lex.clearToken();
	    lex.state = 1;
	    
    	case 1:
	    // Body of comment
    	    if (ch == '*') {
    		lex.state = 2;
	    }
    	    return;
	    
    	case 2:
	    // After *  Complete if we see a /
    	    if (ch == '/') {
		var val = lex.getToken();
//		console.log ("Comment="+val);
		
    		lex.completeTokenValue(val.substr(1, val.length-3));
	    }
    	    else if (ch != '*')
    		lex.state = 1;
    	}
    }

    
    // This is a bit more general than P21 allows.  This also parses the
    // ISO-10303-21 at the beginning and the ANCHOR_NAME    
    function scan_keyword(lex, ch, ctype) {
    	switch (ctype) {
    	case cls.UPPER:
    	case cls.LOWER:
    	case cls.DIGIT:
//    	    lex.kw += ch;
    	    return;
    		
    	case cls.SPECIAL:
    		
    	    if (ch == '-') {
  //  		lex.kw += ch;
    		return;
    	    }
    	}
	
    	lex.completeToken(ch);
    }


    function scan_number(lex, ch, ctype) {

	if (!lex.state) {
	    lex.state = 1;

	    var tok = lex.getToken()
	    if (tok == '+')
		lex.clearToken();
	    else if (get_char_class(tok) == cls.DIGIT)
		lex.state = 2;
	}
    	
    	var done = false;
	
    	switch (lex.state) {
    		
    	case 1:
    	    // Before any digits
    	    if (ctype == cls.DIGIT) {
    		lex.state = 2;
    	    }
    	    else {
    		throw new P21.SyntaxError (
    		    lex,"in number: no integer portion: " + kw);
    	    }
    	    break;
    		
    	case 2:
    	    // Integer portion 
    	    if (ctype == cls.DIGIT)
		;
    	    else if (ch == ".") {
    		lex.state = 3;
    	    }
    	    else
    		done = true;
    	    break;
    		
    	    
    	case 3:
    	    // Fraction
    	    if (ctype == cls.DIGIT)
		;
    	    else if (ch == 'E') {
    		lex.state = 4;
    	    }
    	    else
    		done = true;
    	    break;
    	    
    	case 4:
    	    // Exponent
    	    if (ctype == cls.DIGIT) {
    		lex.state = 6;
    	    }
    	    else if ( ch == '-') {
    		lex.state = 5;
    	    }
    	    else if (ch == '+')
    		lex.state = 5;
    	    else
    		throw new P21.SyntaxError (lex, "no exponent digits");
    	    break;
    	    
    	case 5:
    	    // After sign on exponent, but before first digit 
    	    if (ctype == cls.DIGIT) {
    		lex.state = 6;
    	    }
    	    else
    		throw new P21.SyntaxError (lex, "no exponent digits");
    	    break;
    		
    	case 6:
    	    if (ctype == cls.DIGIT) {
    		;
    	    }
    	    else
    		done = true;
    	    break;
    	    
    	}

	if (done) {
    	
    	    var val;
	    var tok = lex.getToken();

//	    console.log ("Number=" +tok);
//	    console.log ("State=" +lex.state);
	    
    	    if (lex.state == 2) {
    		val = new P21.Integer(tok);
    	    }
    	    else {
    		val = new P21.Real(tok);
    	    }
    	    
    	    lex.completePrimitive(val, ch);
	}
    }
    
    
    
    function scan_instance_name(lex, ch, ctype)
    {
	if (!lex.state) {
	    lex.clearToken();
	    lex.state = 1;
	}

	switch (ctype) {
	case cls.LOWER:
	case cls.UPPER:
	case cls.DIGIT:
	    return;
		
	default:
	    lex.completeTokenValue(new P21.ID(lex.kw), ch);
	}
    }

    function scan_enum(lex, ch, ctype)
    {
	if (!lex.state) {
	    lex.state = 1;
	    lex.clearToken();
	}
    
	switch (lex.state) {
		
	case 1:
	    if (ctype == cls.UPPER) {
		lex.state = 2;
	    }
	    else 
		throw new P21.SyntaxError
	    (lex, "bad enum: expecting upper case character");
	    
	    break;
		
	case 2:
	    if (ctype == cls.UPPER || ctype == cls.DIGIT) {
		lex.state = 2;
	    }
	    else if (ch == '.')
		lex.completePrimitive(new P21.Enum(lex.getToken()));
		
	    else 
		throw new P21.SyntaxError (lex, "in enum");	    
	}
    }
    

    function scan_string(lex, ch, ctype) {
	if (!lex.state) {
	    lex.clearToken();
	    lex.state = 1;
	}

    	switch (lex.state) {
    	case 1:
    	    if (ch == '\'') {
    		lex.state = 2;
    		break;
    	    }
    	    break;
    		
    	case 2:
    	    /* After a quote.  We are either at the end of the string,
    	     * or we have an escaped (doubled) quote
    	     */
    	    if (ch == '\'') {
    		lex.state = 1;
    		break;
    	    }

	    var tok = lex.getToken();
	    var string = tok.substr(0, tok.length-1);
    	    lex.completePrimitive(P21.String.fromP21String(string), ch);
    	}
    }

    
    function scan_binary(lex, ch) {

	if (!lex.state) {
	    lex.state = 1;
	    lex.clearToken();
	}
    	
    	switch (lex.state) {
    	case 1:
    	    if (ch >= '0' && ch <= '3') {
    	    	lex.state = 2;
    	    }
    	    else
    		throw new P21.SyntaxError (
    		    lex, "binary must begin with 0-3");
    	    break;
    		
    	case 2:
    	    if (ch == '"') {
    		lex.completePrimitive(new P21.Binary(lex.getToken()));
	    }
    	    else if (ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F')
    		;
	    else
    		throw new P21.SyntaxError (
    		    lex, "illegal character in binary: "+ch);
    	    break;	    
    	}
    	
    }


    
    function scan_uri(lex, ch) {
	
	if (!lex.state) {
    	    lex.state = 1;
	    lex.clearToken();
	}

	if (ch == '>') {
	    lex.completeTokenValue(new P21.URI(lex.kw));
	    return;
	}
    }
    

    function process_char(lex,ch) {

//	console.log ("Have ch="+ch);
	
	if (ch == '\n')
	    lex.lineno ++;

	
	var ctype = get_char_class(ch);

//	console.log ("Line="+lex.lineno+ " ctype="+ctype);
	
//	return;
	
	// Skip chars out of alphabet
	if (ctype == cls.OOA)
	    return;

	if (!lex.token_state) {
	    lex.kw = ch;
	    lex.state = null;
	    
	    switch (ctype) {

	    case cls.UPPER:
	    case cls.LOWER:
		lex.token_state = cls.TOK_KEYWORD;
		return;
	    
	    case cls.DIGIT:
		lex.token_state = cls.TOK_NUMBER;
		return;
		
	    case cls.APOSTROPHE:
		lex.token_state = cls.TOK_STRING;

	    case cls.SPACE:
		return;
		
	    case cls.SPECIAL:
		switch (ch) {
		    
		case '/':
		    lex.token_state = cls.TOK_COMMENT;
		    return;
		    
		case '"':
		    lex.token_state = cls.TOK_BINARY;
		    return;
		    
		case '+':
		case '-':
		    lex.token_state = cls.TOK_NUMBER;
		    return;

		    
		case '#':
		    lex.token_state = cls.TOK_ID;
		    return;
		    
		case '.':
		    lex.token_state = cls.TOK_ENUM;
		    return;
		    
		case '<':
		    lex.token_state = cls.TOK_URI;
		    return;
		    
		case '$':
		    lex.completePrimitive(null);
		    return;
		    
		case '*':
		    lex.completePrimitive(P21.OMITTED);
		    return;

		case ';':
		case '(':
		case ')':
		case ',':
		case '=':
		case '{':
		case '}':
		case ':':
		    lex.completeTokenType(cls.TOK_PUNCT, ch);
		    return;
		}
		
	    default:
		throw new Error ("Unimplemented token start: "+ctype+"='"
				 +ch+"'");
		return;
		
	    }
	}

	else {
	    switch(lex.token_state) {
		
	    case cls.TOK_KEYWORD:
		scan_keyword(lex, ch, ctype);
		break;

	    case cls.TOK_STRING:
		scan_string(lex, ch, ctype);
		break;

	    case cls.TOK_COMMENT:
		scan_comment(lex, ch);
		break;		

	    case cls.TOK_URI:
		scan_uri(lex, ch);
		break;
		
	    case cls.TOK_ID:
		scan_instance_name(lex, ch, ctype);
		break;

	    case cls.TOK_NUMBER:
		scan_number(lex, ch, ctype);
		break;

	    case cls.TOK_ENUM:
		scan_enum(lex, ch, ctype);
		break;

	    case cls.TOK_BINARY:
		scan_binary(lex, ch, ctype);
		break;
		
	    default:
		throw new P21.SyntaxError (lex, "Unimplemented state: "
					   +lex.token_state);		
	    }

	    if (lex.token_state) {
		lex.kw += ch;
	    }
	}
    }

    cls.prototype.clearToken = function() {
	this.kw = "";
    }

    cls.prototype.getToken = function() {
	return this.kw;
    }
    
    cls.prototype.completeTokenType = function(type, val, push_chars) {
	if (type === undefined) {
	    throw new Error ("Have undefined token type");
	}
	
//	console.log ("Token: "+type+"="+val);

	if (this.processToken) {
	    this.processToken({type: type,
			       value:val,
			       file: this.getFilename(),
			       lineno : this.getLineNumber()
			      });
	}
	
	this.token_state = null;
	this.push_chars = push_chars;
    }

    cls.prototype.completeToken = function(push_chars) {
	this.completeTokenType(this.token_state, this.kw, push_chars);
    }

    cls.prototype.completeTokenValue = function(val, push_chars) {
	this.completeTokenType(this.token_state, val, push_chars);
    }
     
    cls.prototype.completePrimitive = function(val, push_char) {
	this.completeTokenType("primitive", val, push_char);
    }
    
   
    cls.prototype.processChar = function(ch) {	
	process_char(this, ch);

    	while (this.push_chars) {
	    
     	    ch = this.push_chars.charAt(0);
	    
     	    this.push_chars = this.push_chars.substr(1);

     	    process_char(this, ch);
     	}
    }


    cls.prototype.processData = function(str) {
	
	for (var i=0; i<str.length; i++) {
	    var ch = str.charAt(i);
	    this.processChar(ch);
	}
	
    }
    
    cls.prototype.readFile = function(fname) {
	this.stream = open_file_read(fname);	
	
	this.filename = fname;
	
	var str;

	while (true) {
	    str = read_data(this.stream);
	    if (!str) return;
	    try {
		this.processData(str);
	    } catch (e) {
		if (e instanceof P21.Lexer.CancelParse) {
		    return;
		}
		else {
		    throw e;
		}
	    }
	}
    }
    

    cls.prototype.getLineNumber = function() {
	return this.lineno;
    }
    
    cls.prototype.getFilename = function() {
	return this.filename;
    }

    return cls;
})();

P21.Lexer.CancelParse = (function() {

    function cls() {}

    return cls;
})();
    

P21.SyntaxError = (function() {
    function cls(tok, msg) {
	this.name = "P21SyntaxError";

	var fname;
	var lineno;
	
	if (tok instanceof P21.Lexer) {
	    fname = tok.getFilename();
	    lineno = tok.getLineNumber();	    
	} else {
	    fname = tok.file;
	    lineno = tok.lineno;
	}
	
	var message = fname + " line: "+ lineno + ": " + msg;	
	
	this.message = message;

//	return new Error (message);
    }
    
    cls.prototype = SUBTYPE(Error);

    return cls;
})();
