/* $RCSfile: $
 * $Revision: $ $Date: $
 * Auth: Jochen Fritz (jfritz@steptools.com)
 * 
 * Copyright (c) 1991-2013 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 */

"use strict";

P21.Parser = (function() {

    function cls() {
	this.state = 'FILE';
	this.stack = [];
	var self = this;

	this.stack.top = function() {
	    return this[this.length-1];
	}
    }

    function syntax_error(tok, msg) {
//	throw new Error(msg + tok.lineno);
	throw new P21.SyntaxError(tok, msg);
    }

    function expect_punct(self, ch, tok) {
	if (tok.type != P21.Lexer.TOK_PUNCT ||
	    tok.value != ch) 
	    syntax_error(tok, "Expecting punctuation: "+ch+ ", got "+tok.value);
    }
    
    function expect_keyword(self, kw, tok) {
	if (tok.type != P21.Lexer.TOK_KEYWORD||
	    tok.value != kw) 
	    syntax_error(tok, "Expecting keyword: "+kw+ "Got: "+tok.type);
    }

    function expect_type(self, type, tok) {
	if (tok.type != type)
	    syntax_error(tok, "Expecting token of type: "+type+ "; Got: "
			 +tok.type);
    }

    function get_keyword(self, tok) {
	expect_type(self, P21.Lexer.TOK_KEYWORD, tok)
	return tok.value;
    }

    function get_uri(self, tok) {
	expect_type(self, P21.Lexer.TOK_URI, tok)
	return tok.value;
    }

    function get_id(self, tok) {
	expect_type(self, P21.Lexer.TOK_ID, tok)
	return tok.value;
    }
    
    function get_punct(self, tok) {
	expect_type(self, P21.Lexer.TOK_PUNCT, tok)
	return tok.value;
    }
    
//    cls.prototype.getLexer = function() {return this.lexer;}

    
    cls.prototype.token = function(tok) {

//	var tok = {type:type, value:val};
	
//	console.log ("Got token: "+tok.type+
//		     " val="+P21.toP21String(tok.value));

	if (tok.type == P21.Lexer.TOK_COMMENT)
	    return;

	var kw;
	
	switch(this.state) {
	case 'FILE':
	    expect_keyword(this, "ISO-10303-21", tok);
	    this.state = 'FILE_1';
	    break;
	    
	case 'FILE_1':
	    expect_punct(this, ";", tok);
	    this.state = 'FILE_2';
	    break;
	    
	case 'FILE_2':
	    expect_keyword(this, "HEADER", tok);

	    if (this.beginHeaderSection)
		this.beginHeaderSection();
	    
	    this.state = 'FILE_3';
	    break;
	    
	case 'FILE_3':
	    expect_punct(this, ";", tok);
	    this.state = 'HEAD_1';
	    break;

	case 'FILE_4':
	    var kw = get_keyword(this, tok);
	    if (kw == 'ANCHOR') {
		this.state = 'ANCHOR';

		if (this.beginAnchorSection)
		    this.beginAnchorSection();
		
		break;
	    }
	    // Fall-thru if we do not have a match
	    
	case 'FILE_5':
	    var kw = get_keyword(this, tok);
	    if (kw == 'REFERENCE') {
		this.state = 'REFERENCE';

		if (this.beginReferenceSection)
		    this.beginReferenceSection();
		
		break;
	    }
	    // Fall-thru if we do not have a match
	    
	case 'FILE_6':
	    var kw = get_keyword(this, tok);
	    if (kw == 'DATA') {
		this.state = 'DATA';

		if (this.beginDataSection)
		    this.beginDataSection();
		
		break;
	    }
	    else 
		syntax_error(tok, "Unexpected section token: "+tok.type);
	    
	    break;

	    
	    //////////////////////////////////
	    // Header section
	    
	case 'HEAD_1':		
	    this.stack.push({entity: get_keyword(this, tok)}) ;
	    this.state = 'HEAD_2';				
	    break;
	    
	case 'HEAD_2':
	    var p = get_punct(this, tok);
	    if (p == ';') {
		if (this.stack.top().entity != 'ENDSEC')
		    syntax_error(tok, "Expecting ENDSEC");
		this.state = 'FILE_4';

		if (this.endHeaderSection)
		    this.endHeaderSection();
		
	    }
	    
	    else if (p == '(') {
		this.state = 'LIST';
		this.stack.top().cont = function(val) {
		    this.state = 'HEAD_3';

		    var tos = this.stack.pop();
		    var name = tos.entity;

		    
		    if (this.headerEntity)
			this.headerEntity(name, val);
		    // Report list
		}
	    }
	    
	    break;
	    

	case 'HEAD_3':
	    expect_punct(this, ";", tok);
	    this.state = 'HEAD_1';
	    

	    break;


	    //////////////////////////////////////////////////////
	case 'LIST':
	    this.stack.top().listval = [];
	    if (tok.type == P21.Lexer.TOK_PUNCT && tok.value == ')') {
		var tos = this.stack.top();
		tos.cont.call(this, tos.listval);
		break;
	    }
	    
	    // no break 

	case 'LIST_PARAM':
	    switch(tok.type) {

	    case P21.Lexer.TOK_PRIMITIVE:
	    case P21.Lexer.TOK_ID:
		this.stack.top().listval.push(tok.value);
		this.state = 'LIST_AFTER_PARAM';
		break;

	    case P21.Lexer.TOK_KEYWORD:
		this.state = 'TYPED';		
		this.stack.push({
		    type: tok.value,
		    cont: function(v) {
			this.stack.pop();
			this.stack.top().listval.push(v);
			this.state = 'LIST_AFTER_PARAM';
		    }
		});
		
		break;
		
	    case P21.Lexer.TOK_PUNCT:

		if (tok.value == '(') {
		    this.state = 'LIST';
		    this.stack.push({cont: function(val) {
			var tos = this.stack.pop();
			this.stack.top().listval.push(val);
			this.state = 'LIST_AFTER_PARAM';
		    }
				    });
		}
		else 
		    syntax_error(tok, "Unpexected punctuation in paramter: '"
				 +tok.value + "'");
		break;

	    default:
		syntax_error (tok, "Unimplemented token in plist:" +tok.type);
	    }
	    
	    break;


	case 'LIST_AFTER_PARAM':
	    var p = get_punct(this, tok);
	    if (p == ')') {
		var tos = this.stack.top();
		tos.cont.call(this, tos.listval);
		break;
	    }
	    else if (p == ',') {
		this.state = 'LIST_PARAM';
	    }
	    
	    else
		syntax_error (tok, "Unexpected character after param: "+p);

	    break;


	case 'TYPED':
	    expect_punct(this, '(', tok);
	    this.state = 'TYPED_1';
	    break;

	case 'TYPED_1':
	    switch(tok.type) {
	    case P21.Lexer.TOK_PRIMITIVE:
	    case P21.Lexer.TOK_ID:
		this.stack.top().val = tok.value;
		this.state = 'TYPED_2';
		break;

	    case P21.Lexer.TOK_KEYWORD:
		this.state = 'TYPED';		
		this.stack.push({
		    type: tok.value,
		    cont: function(v) {
			this.stack.pop();
			this.stack.top().val = v;
			this.state ='TYPED_2';
		    }
		});
		
		break;
		
	    case P21.Lexer.TOK_PUNCT:

		if (tok.value == '(') {
		    this.state = 'LIST';
		    this.stack.push({
			cont: function(val) {
			    var tos = this.stack.pop();
			    this.stack.top().val = val;
			    this.state = 'TYPED_2';
			}
		    });
		}
		else 
		    syntax_error(tok, "Unpexected punctuation in paramter: '"
				 +tok.value + "'");
		break;
		
	    default:
		syntax_error (tok, "Unimplemented token in typed param:"
			      +tok.type);
	    }
	    break;

	case 'TYPED_2':
	    expect_punct(this, ')', tok);
	    // Report Value
	    var tos = this.stack.top();
	    tos.cont.call(this, tos.val);
	    break;
	    
	    ///////////////////////////////////////////////////

	case 'ANCHOR':
	    expect_punct(this, ';', tok);
	    this.state = 'ANCHOR_1';
	    break;

	case 'ANCHOR_1':
	    if (tok.type == P21.Lexer.TOK_KEYWORD) {
		if (tok.value == 'ENDSEC')
		    this.state = 'ANCHOR_END';
		else
		    syntax_error(tok, "Expected ENDSEC or <anchor>");
	    }

	    else {
		this.stack.push({name: get_uri(this, tok), tags: []}) ;
		this.state = 'ANCHOR_2';
	    }
	    break;

	case 'ANCHOR_2':
	    var p = get_punct(this, tok);
	    if (p == '{') {
		this.state = 'TAG';
		this.stack.top().cont = function(tname, tval) {
		    this.stack.top().tags.push({name: tname, value: tval});
		    // Report tag		    
		    this.state = 'ANCHOR_2';		    
		}
	    }	    
	    else if (p == '=') {
		this.state = 'RESOURCE_LITERAL';
		this.stack.top().cont = function(value) {
		    var tos = this.stack.pop();
		    var name = tos.name;
		    var tags = tos.tags;
		    this.state = 'ANCHOR_AFTER';

		    if (this.anchor) {
			this.anchor(name, value, tags);
		    }
		};
	    }
	    else
		syntax_error(tok, "Unexpected token after anchor :'" + p
			     + "'");
	    
	    break;

	case 'ANCHOR_AFTER':
	    expect_punct(this, ";", tok);
	    this.state = 'ANCHOR_1';
	    break;

	case 'ANCHOR_END':
	    expect_punct(this, ";", tok);
	    this.state = 'FILE_5';

	    if (this.endAnchorSection)
		this.endAnchorSection();
	    
	    break;



	    //////////////////////////////////////////////

	case 'REFERENCE':
	    expect_punct(this, ';', tok);
	    this.state = 'REFERENCE_1';
	    break;

	case 'REFERENCE_1':
	    if (tok.type == P21.Lexer.TOK_KEYWORD) {
		if (tok.value == 'ENDSEC')
		    this.state = 'REFERENCE_END';
		else
		    syntax_error(tok, "Expected ENDSEC or instance name");
	    }

	    else {
		this.stack.push({name: get_id(this, tok), tags: []}) ;
		this.state = 'REFERENCE_2';
	    }
	    break;

	case 'REFERENCE_2':
	    var p = get_punct(this, tok);
	    if (p == '{') {
		this.state = 'TAG';
		this.stack.top().cont = function(tname, tval) {
		    this.stack.top().tags.push({name: tname, value: tval});
		    this.state = 'REFERENCE_2';		    
		}
	    }	    
	    else if (p == '=') {
		this.state = 'RESOURCE_LITERAL';
		this.stack.top().cont = function(value) {
		    var tos = this.stack.pop();
		    var name = tos.name;
		    var tags = tos.tags;
		    
		    this.state = 'REFERENCE_AFTER';

		    if (this.reference) {
			this.reference(name, value, tags);
		    }
		};
		
	    }
	    else
		syntax_error(tok, "Unexpected token after reference :'" + p
			     + "'");
	    
	    break;

	case 'REFERENCE_AFTER':
	    expect_punct(this, ";", tok);
	    this.state = 'REFERENCE_1';
	    break;

	case 'REFERENCE_END':
	    expect_punct(this, ";", tok);
	    this.state = 'FILE_6';

	    if (this.endReferenceSection)
		this.endReferenceSection();
	    
	    break;


	    /////////////////////////////////////////////////////////////

	case 'TAG':
	    this.stack.top().tagname = get_keyword(this, tok);
	    this.state = 'TAG_1';
	    break;

	case 'TAG_1':
	    expect_punct(this, ":", tok);
	    this.state = 'RESOURCE_LITERAL';
	    this.stack.push({cont: function(value){
		this.stack.pop();
		var tos = this.stack.top();
		tos.tagval = value;
		this.state = 'TAG_2';		
	    }});
	    break;

	case 'TAG_2':
	    expect_punct(this, "}", tok);
	    var tos = this.stack.top();
	    this.stack.top().cont.call(this, tos.tagname, tos.tagval);
	    break;
	    
	    /////////////////////////////////////////////////////////////

	case 'RESOURCE_LITERAL':
	    switch (tok.type) {

	    case P21.Lexer.TOK_URI:
	    case P21.Lexer.TOK_PRIMITIVE:
	    case P21.Lexer.TOK_ID:
		this.stack.top().cont.call(this, tok.value);
		break;
		
	    case P21.Lexer.TOK_PUNCT:
		if (tok.value == '(') {
		    this.state = 'LIT_LIST';
		    
		    this.stack.push({cont : function(l) {
			this.stack.pop();
			this.stack.top().cont.call(this, l);
		    }});
		}
		else
		    syntax_error (tok, "Unexpected punctuation token in anchor");
		break;
		
	    default:
		syntax_error (tok, "Unexpected token in anchor default: "
			      + tok.type);
	    }

	    break;


	case 'LIT_LIST':
	    this.stack.top().listval = [];
	    // no break 

	case 'LIT_LIST_PARAM':
	    switch(tok.type) {

	    case P21.Lexer.TOK_PRIMITIVE:
	    case P21.Lexer.TOK_ID:
	    case P21.Lexer.TOK_URI:
		this.stack.top().listval.push(tok.value);
		this.state = 'LIT_LIST_AFTER_PARAM';
		break;
				
	    case P21.Lexer.TOK_PUNCT:

		if (tok.value == '(') {
		    this.state = 'LIT_LIST';
		    this.stack.push({cont: function(val) {
			var tos = this.stack.pop();
			this.stack.top().listval.push(val);
			this.state = 'LIT_LIST_AFTER_PARAM';
		    }});
		}
		else 
		    syntax_error(tok, "Unpexected punctuation in paramter: '"
				 +tok.value + "'");
		break;

	    default:
		syntax_error (tok, "Unimplemented token in literal list:" +tok.type);
	    }
	    
	    break;


	case 'LIT_LIST_AFTER_PARAM':
	    var p = get_punct(this, tok);
	    if (p == ')') {
		var tos = this.stack.top();
		tos.cont.call(this, tos.listval);
		break;
	    }
	    else if (p == ',') {
		this.state = 'LIT_LIST_PARAM';
	    }
	    
	    else
		syntax_error (tok, "Unexpected character after param: "+p);

	    break;


	    /////////////////////////////////////////////////////////

	case 'DATA':
	    var p = get_punct(this, tok);
	    if (p == '(') {
		this.state = 'LIST';
		this.stack.top().cont = function(val) {
		    this.state = 'DATA_1';
		    // Report list
		}		
	    }
	    else if (p == ';')
		this.state = 'INSTANCES';
	    break;

	case 'DATA_1':
	    expect_punct(this, ';', tok);
	    this.state = 'DATA_2';
	    break;

	case 'INSTANCES':
	    if (tok.type == P21.Lexer.TOK_KEYWORD) {
		if (tok.value == 'ENDSEC') 
		    this.state = 'END';
		else
		    syntax_error (tok, "Unexpected keyword in data section");
	    }
	    else if (tok.type == P21.Lexer.TOK_ID) {
		this.stack.top().eid = tok.value;
		this.state = 'INSTANCES_1';
	    }
	    else
		syntax_error (tok, "Unexpected token in data section");
	    break;

	case 'INSTANCES_1':
	    expect_punct(this, '=', tok);
	    this.state = 'INSTANCES_2';
	    break;

	case 'INSTANCES_2':
	    if (tok.type == P21.Lexer.TOK_PUNCT && tok.value == '(') {		
		this.state = 'COMPLEX';
		this.stack.top().complex = {};
		break;
	    }
	    else if (tok.type == P21.Lexer.TOK_KEYWORD) {
		this.stack.top().entity = tok.value;
		this.state = 'INSTANCES_3';
		break;
	    }
	    else
		syntax_error (tok, "Expecteded token in entity");
	    break;


	case 'INSTANCES_3':
	    expect_punct(this, '(', tok);
	    this.state = 'LIST';
	    this.stack.top().cont = function(val) {
		var tos = this.stack.top();		
		// Report instance complete

		if (this.instance) {
		    this.instance(tos.eid, tos.entity, val);
		}

		delete tos.eid;
		delete tos.entity;
		
		this.state = 'INSTANCES_4';
	    }
	    break;

	case 'INSTANCES_4':
	    expect_punct(this, ";", tok);
	    this.state = 'INSTANCES';
	    break;


	case 'COMPLEX':
	    if (tok.type == P21.Lexer.TOK_PUNCT && tok.value == ')') {
		// Report complex instance
		var tos = this.stack.top();
		
		if (this.complex)
		    this.complex(tos.eid, tos.complex);

		delete tos.eid;
		delete tos.complex;
		
		this.state = 'INSTANCES_4';		
	    }

	    else {
		var kw = get_keyword(this, tok);
		this.stack.push({entity: kw});
		this.state = 'COMPLEX_1';
	    }
	    break;

	case 'COMPLEX_1':
	    expect_punct(this, '(', tok);
	    this.state = 'LIST';
	    this.stack.top().cont = function(val) {
		var tos = this.stack.pop();
		// Save instance record
		var entity = tos.entity;

		this.stack.top().complex[entity] = val;
		
		this.state = 'COMPLEX';
	    }
	    break;
	    
	    
	case 'END':
	    expect_punct(this, ';', tok)
	    this.state = 'END_1';

	    if (this.endDataSection)
		this.endDataSection();
	    
	    break;


	case 'END_1':
	    expect_keyword(this, 'END-ISO-10303-21', tok);
	    this.state = 'END_2';
	    break;
	    
	case 'END_2':
	    expect_punct(this, ';', tok);
	    this.state = null;

	    if (this.endFile)
		this.endFile();
	    
	    break;
	    
	default:
	    syntax_error (tok, "Unimplemented state: " + this.state);
	    
	}
    }

    
    return cls;
    
}());
