/* $RCSfile: $
 * $Revision: $ $Date: $
 * Auth: Jochen Fritz (jfritz@steptools.com)
 * 
 * 	Copyright (c) 1991-2013 by 
 * 	STEP Tools Inc., Troy, New York
 * 	All Rights Reserved
 * 
 * 	This software is furnished under a license and may be used and
 * 	copied only in accordance with the terms of such license and with
 * 	the inclusion of the above copyright notice.  This software and
 * 	accompanying written materials or any other copies thereof may
 * 	not be provided or otherwise made available to any other person.
 * 	No title to or ownership of the software is hereby transferred.
 * 
 * 		----------------------------------------
 */



/*
 * Load the P21 Interface for the Node.js environment
 * This is not a node.js module.  Instead, we are using the traditional
 * method of managinf javascript files, so that the source code is portable
 * to other environments.
 */

var fs = require('fs');
var vm = require('vm');


function include(path) {
    var code = fs.readFileSync(path);

    try {
	vm.runInThisContext(code, path);
    } catch (ex) {
	console.log("Syntax error in include file: " +path);

	if (! path.indexOf("/") > -1)
	    path = "./" + path;

	// We have a syntax error, but we do cannot get a usable back trace.
	// (the trace will point to the vm.runInThisContext, and not the actual
	// file).  Thus, we use the node.js require function to get a better
	// trace.
	// Run-time errors (not parse time) do give us a good trace.
	// Hopefully node will be updated sometime, and this hack will no longer
	// be required.
	require (path);

	// Require should have thrown a syntax error, but if not, we rethrow
	// the original exception
	throw ex;
    }
}

global.include = include;




// File I/O portability functions.  They are low-level routines that the P21
// parser needs to have available.
function open_file_read(filename) {

    return {
	fd : fs.openSync(filename, "r"),
//	lineno : 1,
    };    
}

function close_file_read(stream) {

    fs.closeSync(stream.fd);
}


// Return the codepoint of the next character
// function getc(stream) {

//     // if (stream.pushback) {
	
//     // 	var ret = stream.pushback;
//     // 	stream.pushback = null;
//     // 	if (ret == '\n')
//     // 	    stream.lineno++;
	
//     // 	return ret;
//     // }
    
//     if (!stream.string) {
// 	var buff = new Buffer(1024);	
// 	var count = fs.readSync(stream.fd, buff, 0, buff.length);
// 	if (count == 0)
// 	    return 0;  //EOF 
// 	stream.string = buff.toString('utf8', 0, count);
// 	stream.idx = 0;
//     }

//     if (!stream.string)
// 	return 0;
    
//     var ret = stream.string.charAt(stream.idx++);

//     if (stream.idx >= stream.string.length) {
// 	stream.string = null;
//     }
    
//     if (ret == '\n')
// 	stream.lineno++;

//     return ret;
// }


function read_data(stream, sz) {
    if (!sz)
	sz = 10240;

    var buff = new Buffer(sz);
    var count = fs.readSync(stream.fd, buff, 0, buff.length)
    if (count == 0)
	return null;

    var ret = buff.toString('utf8', 0, count);
    return ret;
}


// function ungetc(stream, ch) {
    
//     if (stream.pushback) {
// 	throw new Error ("Already pushed back char");
//     }
//     if (ch == '\n')
// 	stream.lineno--;
    
//     stream.pushback = ch;
// }


function open_file_write(filename) {
    return {
	fd : fs.openSync(filename, "w"),
    };            
}


function close_file_write(stream) {

    fs.closeSync(stream.fd);
}

function write(stream, str) {
    var buff = new Buffer(str);
    fs.writeSync(stream.fd, buff, 0, buff.length);
}

global.open_file_read = open_file_read;
global.open_file_write = open_file_write;
global.close_file_read = close_file_read;
global.close_file_write = close_file_write;
//global.getc = getc;
//global.ungetc = ungetc;
global.read_data = read_data;

global.write_file = write;
global.rename_file = function(oldp, newp) {
    fs.renameSync(oldp, newp);
}

//include("P21Parse.js");
//include("p21.js");


include ("P21Wrapper.js");
include ("P21Lexer.js");
include ("P21Parse.js");
include ("P21Model.js");
include ("P21Reader.js");


// Call the main (user-supplied) code
global.ARGV = process.argv.slice(2);
include(ARGV[0]);

